"""
morinkit.vcf
~~~~~~~~~~~~

This module contains utility functions relevant to VCF files.
"""

from collections import defaultdict, OrderedDict
from cyvcf2 import VCF
from cyvcf2.cyvcf2 import INFO as cyvcf2_INFO


def iter_vcfs(vcf_filenames, is_sorted=False):
    """Load VCF files and return an iterator for each variant.

    Args:
        vcf_files: List of VCF file paths.
        is_sorted: Boolean indicating whether the VCF files are sorted by
            chrom, pos, ref and alt.

    Returns:
        Iterator that yields a list of VCF records for each variant in the
        same order as the `vcf_files` argument.
    """
    vcf_readers = [VCF(vcf) for vcf in vcf_filenames]
    if is_sorted:
        iterator = create_vcf_walktogether(vcf_readers)
    else:
        iterator = create_vcf_iterator(vcf_readers)
    return iterator


def get_key(vcf_record):
    """Get unique variant key for a VCF record.

    Args:
        vcf_record: cyvcf2 VCF record.

    Returns:
        Tuple of VCF attributes.
    """
    return (vcf_record.CHROM, vcf_record.POS, vcf_record.REF, vcf_record.ALT[0])


def create_vcf_iterator(vcf_readers):
    """Create a VcfIterator for sorted and unsorted VCF files.

    This function loads entire files into memory.

    Args:
        vcf_readers: List of cyvcf2 VCF readers.

    Yields:
        List of VCF records with the same length as the number of input VCF
        readers. Yields one list per variant.
    """
    index = defaultdict(lambda: [None] * len(vcf_readers))
    for i, vcf in enumerate(vcf_readers):
        for record in vcf:
            key = get_key(record)
            index[key][i] = record
    for key, val in index.items():
        yield val


def create_vcf_walktogether(vcf_readers):
    """Create a VCF iterator that efficiently handles sorted VCF files.

    Behaviour on unsorted files is undefined. Adapted from PyVCF walk_together:
    https://github.com/jamescasbon/PyVCF/blob/master/vcf/utils.py

    Args:
        vcf_readers: List of cyvcf2 VCF readers

    Yields:
        List of VCF records with the same length as the number of input VCF
        readers. Yields one list per variant.
    """
    def next_record(vcf_reader):
        try:
            next_record = next(vcf_reader)
        except StopIteration:
            next_record = None
        finally:
            return next_record

    nexts = []
    for vcf in vcf_readers:
        nexts.append(next_record(vcf))

    min_k = (None,)   #
    while any([r is not None for r in nexts]):

        next_idx_to_k = {}
        for i, record in enumerate(nexts):
            if record is not None:
                next_idx_to_k[i] = get_key(record)

        keys_with_prev_contig = []
        for k in next_idx_to_k.values():
            if k[0] == min_k[0]:
                keys_with_prev_contig.append(k)

        if any(keys_with_prev_contig):
            min_k = min(keys_with_prev_contig)
        else:
            min_k = min(next_idx_to_k.values())

        min_k_idxs = set([i for i, k in next_idx_to_k.items() if k == min_k])

        min_k_idxs = set()
        for i, k in next_idx_to_k.items():
            if k == min_k:
                min_k_idxs.add(i)

        records = []
        for i in range(len(nexts)):
            if i in min_k_idxs:
                records.append(nexts[i])
            else:
                records.append(None)
        yield records

        for i in min_k_idxs:
            nexts[i] = next_record(vcf_readers[i])


class OutputVCF:
    """Utility class for outputting VCF file.

    Limitations:
        Doesn't support FORMAT and sample columns (yet).
    """

    COLS = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO']
    HEADER = '##fileformat=VCFv4.2\n#{}\n'.format('\t'.join(COLS))

    def __init__(self, file=None, filename=None, header=None):
        if file:
            if 'w' not in file.mode:
                raise TypeError('`file` mode not set to writeable.')
            self.file = file
        elif filename:
            self.file = open(filename, 'w')
        else:
            raise TypeError('Both `file` and `filename` arguments are undefined.')
        self.header = header or self.HEADER
        self.write_header()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.file.close()

    @classmethod
    def info_to_string(cls, key, val):
        if val is None:
            result = key
        else:
            result = '{}={}'.format(key, cls.to_string(val))
        return result

    @classmethod
    def to_string(cls, field):
        if isinstance(field, str):
            result = field
        elif isinstance(field, bytes):
            result = str(field, 'utf-8')
        elif field is None:
            result = '.'
        elif isinstance(field, cyvcf2_INFO):
            infos = [cls.info_to_string(key, val) for key, val in field]
            result = ';'.join(infos)
        elif isinstance(field, list):
            result = ','.join(field)
        elif isinstance(field, int):
            result = str(field)
        elif isinstance(field, float):
            result = str(round(field, 2))
        else:
            raise ValueError('Unsupported field: {}'.format(field))
        return result

    def write_header(self):
        """Write header to VCF file.

        Returns:
            None
        """
        self.file.write(self.header)

    def write_record(self, record, **kwargs):
        """Write VCF row to file based on a given VCF record.

        Args:
            record: cyvcf2 VCF record.
            kwargs: Dictionary of VCF columns (keys) to be overwritten with a
                given string (values).

        Returns:
            None
        """
        row = OrderedDict((col, self.to_string(getattr(record, col))) for col in self.COLS)
        # Update any columns based on kwargs
        for col in kwargs.keys() & self.COLS:
            row[col] = kwargs[col]
        line = '\t'.join(row.values()) + '\n'
        self.file.write(line)

    def write_row(self, CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO):
        cols = [CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO]
        line = '\t'.join(cols) + '\n'
        self.file.write(line)



"""
morinkit.morinkit
~~~~~~~~~~~~~~~~~

This module acts as the entry point to the morinkit package.
"""

import sys
import logging
from .ui import parse_args
from .utils import config_logging
from . import convert, process, summarize
from .exceptions import NotUniqueError

MODULES = [convert, process, summarize]


def find_func(name, modules=MODULES):
    """Find function with a given name in a given set of modules.

    This is mainly used to find the relevant function for a given
    subcommand.

    Args:
        name (str): Function name.
        modules (list): List of modules.

    Returns:
        Function object or None.

    Raises:
        NotUniqueError: Found more than one function with the given name
    """
    assert isinstance(name, str), '`name` is not a string'
    assert isinstance(modules, list), '`modules` is not a list'
    funcs = []
    for m in modules:
        func = getattr(m, name, None)
        if func is not None:
            funcs.append(func)
    if len(funcs) == 0:
        return None
    if len(funcs) == 1:
        return funcs[0]
    elif len(funcs) > 1:
        raise NotUniqueError('Found more than one function named {}'.format(name))


def main():
    """Call relevant function based on command-line arguments."""
    config_logging()
    logging.info('Command: {}'.format(' '.join(sys.argv)))
    args = parse_args()
    subcommand = args.pop("subcommand")
    func = find_func(subcommand)
    func(**args)


if __name__ == '__main__':
    main()



"""
morinkit.process
~~~~~~~~~~~~~~~~

This module contains top-level functions that processes files without changing
the file format.
"""


from .vcf import iter_vcfs, OutputVCF


def ensemble_vcf(vcf_filenames, output_file, is_sorted, min_support):
    """Integrate the variant calls from multiple VCF files and output an
    ensemble VCF file.

    ensemble_vcf is primarily meant for ensemble SNV calling. In other words,
    this function creates a VCF file where each variant is supported by at
    least `min_support` methods.

    Args:
        vcf_files: List of VCF file paths.
        output: Opened writeable file.
        is_sorted: Boolean indicating whether the input VCF files are sorted.
        min_support: Integer or fraction indicating the minimum number of
            methods supporting each variant.

    Returns:
        None
    """
    overwrite = {'ID': '.', 'QUAL': '.', 'FILTER': '.', 'INFO': '.'}
    num_methods = len(vcf_filenames)
    with OutputVCF(file=output_file) as ovcf:
        for all_records in iter_vcfs(vcf_filenames, is_sorted=is_sorted):
            # filter(None, records) removes items that evaluate to False (incl. None)
            # Because the records represent the same variant, we take the first one
            records = list(filter(None, all_records))
            num_called = len(records)
            # If min_support is less than 1, consider the fraction of methods
            # supporting the variant.
            if min_support < 1 and num_called / num_methods < min_support:
                continue
            # If the min_support is greater or equal to one, consider the
            # absolute number of methods supporting the variant.
            elif min_support >= 1 and num_called < min_support:
                continue
            # Write out record while overwritting unnecessary VCF fields.
            ovcf.write_record(records[0], **overwrite)



"""
morinkit
~~~~~~~~

This package contains Python functions and script for genomics.
"""

__title__ = 'morinkit'
__version__ = '0.1.1'
__author__ = 'Morin Lab Members (Bruno Grande)'
__license__ = 'GNU GPLv3'
__url__ = 'https://github.com/morinlab/morinkit'



"""
morinkit.ui
~~~~~~~~~~~

This module defines the CLI for calling various top-level functions.
"""

import sys
from argparse import ArgumentParser, FileType


def parse_args():
    """Parse command-line arguments.

    Returns:
        A dict containing the subcommand and all other positional and
        optional arguments.
    """
    # Initialize parser and subparsers
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest='subcommand')

    # ensemble_vcf
    ensemble_vcf = subparsers.add_parser('ensemble_vcf')
    ensemble_vcf.add_argument('vcf_filenames', nargs='+', metavar='vcf_file')
    ensemble_vcf.add_argument('--output', '-o', type=FileType('w'), default=sys.stdout,
                              dest='output_file')
    ensemble_vcf.add_argument('--sorted', action='store_true', dest='is_sorted',
                              help='Are all input VCF files sorted in the same order? '
                              'N.B. If not sorted, the entire files are loaded into memory.')
    ensemble_vcf.add_argument('--min_support', type=float, default=0.5,
                              help='Minimum number of supporting methods (one method per input '
                              'VCF). If less than 1, minimum fraction of supporting methods.')

    # Parse arguments and return dict
    args = parser.parse_args()
    return vars(args)



#!/usr/bin/env python3

import os
from setuptools import setup

from morinkit import __version__, __author__, __license__, __url__


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

long_desc = 'For more information, visit the GitHub repository: {}'.format(__url__)

requires = read('requirements.txt').rstrip().split("\n")
tests_requires = read('requirements-tests.txt').rstrip().split("\n")


setup(
    name='morinkit',
    version=__version__,
    description='Python functions for cancer genomics',
    long_description=long_desc,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics'],
    keywords='bioinformatics cancer genomics utility tool',
    author=__author__,
    author_email='morinlab@sfu.ca',
    url=__url__,
    license=__license__,
    packages=['morinkit'],
    entry_points={'console_scripts': ['morinkit = morinkit.morinkit:main']},
    install_requires=requires,
    test_suite='tests',
    tests_require=tests_requires)



"""
morinkit.exceptions
~~~~~~~~~~~~~~~~~~~

This module contains exceptions for morinkit.
"""


class Error(Exception):
    """Base error"""


class NotUniqueError(Error):
    """More than one items found when only one is expected."""



"""
morinkit.utils
~~~~~~~~~~~~~~

This module contains general-purpose utility functions.
"""

import sys
import logging


def config_logging():
    """Configure logging."""
    log_format = '%(asctime)s - %(levelname)s (%(module)s.%(funcName)s):  %(message)s'
    date_format = '%Y/%m/%d %H:%M:%S'  # 2010/12/12 13:46:36
    logging.basicConfig(format=log_format, level=logging.INFO,
                        datefmt=date_format, stream=sys.stderr)
